from types import *
import sys, new, string
import template

# ct981218 added missing consts
from template import CO_OPTIMIZED, CO_NEWLOCALS, CO_VARARGS, CO_VARKEYWORDS


#
# XXX Tough call here... should an exception be raised if a keyword
# was specified both when the functor was created and when it was
# called.  Or should the one set at creation be consider a "default"
# to be overridden by the actual call.  Since speed is the motivation
# here, I am just slamming the keyword arg given at creation time on
# top of the one given at call time.
#    


#
# Lazy apply?
#
def xapply(func, *args, **kws):
    return functor(func, args, kws)
    
#
# Given a callable object and argument list and keyword arguments, a
# callable argument will be returned.  Arguments passed in at creation
# time take precedence over those passed at call time.  
#
def functor(func, args, kws):
    if not callable(func):
        raise TypeError, "callable expected"
    if type(func) is FunctionType:
        return _function_functor(func, args, kws)
    if type(func) is MethodType:
        return _instance_method_functor(func, args, kws)
    if type(func) is InstanceType:
        return _instance_functor(func, args, kws)
    return _callable_functor(func, args, kws)


#
# There is a pretty big win on this one
#
def _instance_method_functor(func, args, kws):
    if type(func) is not MethodType:
        raise TypeError, "%s expected" % MethodType.__name__
    return _function_functor(func.im_func, (func.im_self,) + args, kws)


#
# Even bigger here.
#
def _instance_functor(func, args, kws):
    if not callable(func):
        raise TypeError, "callable expected"
    if type(func) is not InstanceType:
        raise TypeError, "%s expected" % InstanceType.__name__
    return _instance_method_functor(func.__call__, args, kws)
    

#
# For a generic callable argument, we can't generate very good
# code since we don't know if it allows keywords or variable
# length arguments list.  All we know is if we have keywords or
# not.  
#
def _callable_functor(func, args, kws):
    try:
        name = func.__name__ + '_functor' 
    except:
        name = 'functor'
    if kws:
        return template.patch_consts(functor_ahk_template, name,
                                        None, locals())
    elif args:
        items = tuple(kws.items())
        return template.patch_consts(functor_ak_template, name,
                                        None, locals())
    return func
    

#    
# There are 5 different case that I thought might be worth dealing
# with.  The choice of which template to use is based on three things:
# 1) did creation supply all the arguments that the function would
# ever need?  2) did creation supply any keywords that could not
# be appended to the argument list? 3) does the function accept
# arbitrary keyword arguments.
#
# A few more cases could be added asking the question: did creation
# supply an argument tuple?  But I think this is overkill already.
#

def _function_functor(func, args, kws):
    if type(func) is not FunctionType:
        raise TypeError, "%s expected" % FunctionType.__name__
    name = func.func_name + '_functor'
    code = func.func_code
    args, kws = template.process_keywords(func, args, kws)
    abbrv = ''
    want_kws = 0
    # need more args?
    if len(args) < code.co_argcount or code.co_flags & CO_VARARGS:
        want_kws = 1
        abbrv = 'a'
    # have keywords?
    if kws:
        want_kws = 1
        items = kws.items()
        abbrv = abbrv + 'h'
    # want more keywords?
    if want_kws or code.co_flags & CO_VARKEYWORDS:
        abbrv = abbrv + 'k'
    functor_template = globals()['_functor_%s_template' % abbrv]
# ct981218 typo corrected:
#   return template.patch_consts(template, name, None, locals())
    return template.patch_consts(functor_template, name, None, locals())


#
# create: all args, no kws
#   call: no args, no kws
#
def _functor__template():
    return '__apply__'('__func__', '__args__')

#
# create: all args, no kws
#   call: no args, more kws
#
def _functor_k_template(**kws):
    return '__apply__'('__func__', '__args__', kws)
    
#    
# create: all args, have kws
#   call: no args, more kws
#
def _functor_hk_template(**kws):
    if kws:
        for k, v in '__items__':
            kws[k] = v
    else:
        kws = '__kws__'
    return '__apply__'('__func__', '__args__', kws)
    
#    
# create: some args, no kws
#   call: more args, more kws
#
def _functor_ak_template(*args, **kws):
    return '__apply__'('__func__', '__args__' + args, kws)
    
#    
# create: some args, have kws
#   call: more args, more kws
#
def _functor_ahk_template(*args, **kws):
    if kws:
        for k, v in '__items__':
            kws[k] = v
    else:
        kws = "__kws__"
    return '__apply__'('__func__', '__args__' + args, kws)
